@@ -25,34 +25,47 @@ from tornado.httputil import HTTPHeaders
 import sys
 import types
 import pandas as pd
+import numpy as np
 from botocore.client import Config
 import ibm_boto3
 from datetime import datetime
+import pyarrow
+import os
+import tempfile
 
 
 class SQLQuery():
-    def __init__(self, api_key, instance_crn, target_cos_url, client_info=''):
+    def __init__(self, api_key, instance_crn, target_cos_url=None, client_info=''):
         self.endpoint_alias_mapping = {
             "us-geo": "s3-api.us-geo.objectstorage.softlayer.net",
-            "dal": "s3-api.dal-us-geo.objectstorage.softlayer.net",
-            "wdc": "s3-api.wdc-us-geo.objectstorage.softlayer.net",
-            "sjc": "s3-api.sjc-us-geo.objectstorage.softlayer.net",
+            "dal-us-geo": "s3-api.dal-us-geo.objectstorage.softlayer.net",
+            "wdc-us-geo": "s3-api.wdc-us-geo.objectstorage.softlayer.net",
+            "sjc-us-geo": "s3-api.sjc-us-geo.objectstorage.softlayer.net",
             "eu-geo": "s3.eu-geo.objectstorage.softlayer.net",
-            "ams": "s3.ams-eu-geo.objectstorage.softlayer.net",
-            "fra": "s3.fra-eu-geo.objectstorage.softlayer.net",
-            "mil": "s3.mil-eu-geo.objectstorage.softlayer.net",
+            "ams-eu-geo": "s3.ams-eu-geo.objectstorage.softlayer.net",
+            "fra-eu-geo": "s3.fra-eu-geo.objectstorage.softlayer.net",
+            "mil-eu-geo": "s3.mil-eu-geo.objectstorage.softlayer.net",
             "us-south": "s3.us-south.objectstorage.softlayer.net",
-            "us-east": "s3.us-east.objectstorage.softlayer.net"
+            "us-east": "s3.us-east.objectstorage.softlayer.net",
+            "jp-tok": "s3.jp-tok.objectstorage.softlayer.net",
+            "ap-geo": "s3.ap-geo.objectstorage.softlayer.net",
+            "tok-ap-geo": "s3.tok-ap-geo.objectstorage.softlayer.net",
+            "seo-ap-geo": "s3.seo-ap-geo.objectstorage.softlayer.net",
+            "hkg-ap-geo": "s3.hkg-ap-geo.objectstorage.softlayer.net",
+            "eu-de": "s3.eu-de.objectstorage.softlayer.net",
+            "eu-gb": "s3.eu-gb.objectstorage.softlayer.net",
+            "ams03": "s3.ams03.objectstorage.softlayer.net",
+            "che01": "s3.che01.objectstorage.softlayer.net",
+            "mel01": "s3.mel01.objectstorage.softlayer.net",
+            "tor01": "s3.tor01.objectstorage.softlayer.net",
+            "osl01": "s3.osl01.objectstorage.softlayer.net",
+            "sao01": "s3.sao01.objectstorage.softlayer.net",
+            "seo01": "s3.seo01.objectstorage.softlayer.net"
         }
         self.api_key = api_key
         self.instance_crn = instance_crn
         self.target_cos = target_cos_url
-        provided_cos_endpoint = target_cos_url.split("/")[2]
-        self.target_cos_endpoint = self.endpoint_alias_mapping.get(provided_cos_endpoint, provided_cos_endpoint)
-        self.target_cos_bucket = target_cos_url.split("/")[3]
-        self.target_cos_prefix = target_cos_url[target_cos_url.replace('/', 'X', 3).find('/')+1:]
-        if self.target_cos_endpoint == '' or self.target_cos_bucket == '' or self.target_cos_prefix == target_cos_url:
-            raise ValueError("target_cos_url value is \'{}\'. Expecting format cos://<endpoint>/<bucket>/[<prefix>]. ".format(target_cos_url))
+        self.export_cos_url = target_cos_url
         if client_info == '':
             self.user_agent = 'IBM Cloud SQL Query Python SDK'
         else:
@@ -96,12 +109,13 @@ class SQLQuery():
         if not self.logged_on:
             print("You are not logged on to IBM Cloud")
             return
-        sqlData = {'statement': sql_text,
-                   'resultset_target': self.target_cos}
+        sqlData = {'statement': sql_text}
+        if self.target_cos:
+            sqlData.update({'resultset_target': self.target_cos})
 
         try:
             response = self.client.fetch(
-                "https://sql-api.ng.bluemix.net/v2-beta/sql_jobs?instance_crn={}".format(self.instance_crn),
+                "https://sql-api.ng.bluemix.net/v2/sql_jobs?instance_crn={}".format(self.instance_crn),
                 method='POST',
                 headers=self.request_headers,
                 validate_cert=False,
@@ -118,7 +132,7 @@ class SQLQuery():
 
         while True:
             response = self.client.fetch(
-                "https://sql-api.ng.bluemix.net/v2-beta/sql_jobs/{}?instance_crn={}".format(jobId, self.instance_crn),
+                "https://sql-api.ng.bluemix.net/v2/sql_jobs/{}?instance_crn={}".format(jobId, self.instance_crn),
                 method='GET',
                 headers=self.request_headers,
                 validate_cert=False)
@@ -147,20 +161,23 @@ class SQLQuery():
             print("You are not logged on to IBM Cloud")
             return
 
-        job_status = self.get_job(jobId)['status']
+        job_details = self.get_job(jobId)
+        job_status = job_details['status']
         if job_status == 'running':
             raise ValueError('SQL job with jobId {} still running. Come back later.')
         elif job_status != 'completed':
             raise ValueError('SQL job with jobId {} did not finish successfully. No result available.')
 
-        if self.target_cos_prefix != '' and not self.target_cos_prefix.endswith('/'):
-            result_location = "https://{}/{}?prefix={}/jobid={}/part".format(self.target_cos_endpoint,
-                                                                        self.target_cos_bucket, self.target_cos_prefix,
-                                                                        jobId)
-        else:
-            result_location = "https://{}/{}?prefix={}jobid={}/part".format(self.target_cos_endpoint,
-                                                                        self.target_cos_bucket, self.target_cos_prefix,
-                                                                        jobId)
+        result_cos_url = job_details['resultset_location']
+        provided_cos_endpoint = result_cos_url.split("/")[2]
+        result_cos_endpoint = self.endpoint_alias_mapping.get(provided_cos_endpoint, provided_cos_endpoint)
+        result_cos_bucket = result_cos_url.split("/")[3]
+        result_cos_prefix = result_cos_url[result_cos_url.replace('/', 'X', 3).find('/')+1:]
+        result_location = "https://{}/{}?prefix={}".format(result_cos_endpoint, result_cos_bucket, result_cos_prefix)
+        result_format = job_details['resultset_format']
+
+        if result_format not in ["csv", "parquet"]:
+            raise ValueError("Result object format {} currently not supported by get_result().".format(result_format))
 
         response = self.client.fetch(
             result_location,
@@ -171,28 +188,111 @@ class SQLQuery():
         if response.code == 200 or response.code == 201:
             ns = {'s3': 'http://s3.amazonaws.com/doc/2006-03-01/'}
             responseBodyXMLroot = ET.fromstring(response.body)
+            bucket_objects = []
+            # Find result objects with data
             for contents in responseBodyXMLroot.findall('s3:Contents', ns):
                 key = contents.find('s3:Key', ns)
-                result_object = key.text
-                # print("Job result for {} stored at: {}".format(jobId, result_object))
+                if int(contents.find('s3:Size', ns).text) > 0:
+                    bucket_objects.append(key.text)
+                #print("Job result for {} stored at: {}".format(jobId, result_object))
         else:
-            print("Result object listing for job {} at {} failed with http code {}".format(jobId, result_location,
+            raise ValueError("Result object listing for job {} at {} failed with http code {}".format(jobId, result_location,
                                                                                            response.code))
 
         cos_client = ibm_boto3.client(service_name='s3',
                                       ibm_api_key_id=self.api_key,
                                       ibm_auth_endpoint="https://iam.ng.bluemix.net/oidc/token",
                                       config=Config(signature_version='oauth'),
-                                      endpoint_url='https://' + self.target_cos_endpoint)
+                                      endpoint_url='https://' + result_cos_endpoint)
 
-        body = cos_client.get_object(Bucket=self.target_cos_bucket, Key=result_object)['Body']
-        # add missing __iter__ method, so pandas accepts body as file-like object
-        if not hasattr(body, "__iter__"): body.__iter__ = types.MethodType(self.__iter__, body)
+        # Loop over result objects and read and concatenate them into result data frame
+        for bucket_object in bucket_objects:
 
-        result_df = pd.read_csv(body)
+            if result_format == "csv":
+                body = cos_client.get_object(Bucket=result_cos_bucket, Key=bucket_object)['Body']
+                # add missing __iter__ method, so pandas accepts body as file-like object
+                if not hasattr(body, "__iter__"): body.__iter__ = types.MethodType(self.__iter__, body)
+
+                partition_df = pd.read_csv(body)
+
+            elif result_format == "parquet":
+                tmpfile = tempfile.NamedTemporaryFile()
+                tempfilename = tmpfile.name
+                cos_client.download_file(Bucket=result_cos_bucket, Key=bucket_object, Filename=tempfilename)
+
+                partition_df = pd.read_parquet(tempfilename)
+                tmpfile.close()
+
+            # Add columns from hive style partition naming schema
+            hive_partition_candidates = bucket_object.replace(result_cos_prefix + '/', '').split('/')
+            for hive_partition_candidate in hive_partition_candidates:
+                if hive_partition_candidate.count('=') == 1: # Hive style folder names contain exactly one '='
+                    column = hive_partition_candidate.split('=')
+                    column_name = column[0]
+                    column_value = column[1]
+                    if column_value == '__HIVE_DEFAULT_PARTITION__': # Null value partition
+                        column_value = np.nan
+                    if len(column_name) > 0 and len(column_value) > 0:
+                        partition_df[column_name] = column_value
+
+            if 'result_df' not in locals():
+                result_df = partition_df
+            else:
+                result_df = result_df.append(partition_df)
+
+        if 'result_df' not in locals():
+            return None
 
         return result_df
 
+    def list_results(self, jobId):
+        if not self.logged_on:
+            print("You are not logged on to IBM Cloud")
+            return
+
+        job_details = self.get_job(jobId)
+        if job_details['status'] == 'running':
+            raise ValueError('SQL job with jobId {} still running. Come back later.')
+        elif job_details['status'] != 'completed':
+            raise ValueError('SQL job with jobId {} did not finish successfully. No result available.')
+
+        result_location = job_details['resultset_location'].replace("cos", "https", 1)
+        provided_cos_endpoint = job_details['resultset_location'].split("/")[2]
+        result_cos_endpoint = self.endpoint_alias_mapping.get(provided_cos_endpoint, provided_cos_endpoint)
+
+        fourth_slash = result_location.replace('/', 'X', 3).find('/')
+
+        response = self.client.fetch(
+            result_location[:fourth_slash] + '?prefix=' + result_location[fourth_slash + 1:],
+            method='GET',
+            headers=self.request_headers,
+            validate_cert=False)
+
+        if response.code == 200 or response.code == 201:
+            ns = {'s3': 'http://s3.amazonaws.com/doc/2006-03-01/'}
+            responseBodyXMLroot = ET.fromstring(response.body)
+            bucket_name = responseBodyXMLroot.find('s3:Name', ns).text
+            bucket_objects = []
+            if responseBodyXMLroot.findall('s3:Contents', ns):
+                for contents in responseBodyXMLroot.findall('s3:Contents', ns):
+                    key = contents.find('s3:Key', ns)
+                    object_url = "cos://{}/{}/{}".format(result_cos_endpoint, bucket_name, key.text)
+                    size = contents.find('s3:Size', ns)
+                    bucket_objects.append({'Key': object_url, 'Size': size.text})
+            else:
+                print('There are no result objects for the jobid {}'.format(jobId))
+                return
+        else:
+            print("Result object listing for job {} at {} failed with http code {}".format(jobId, result_location,
+                                                                                           response.code))
+            return
+
+        result_objects_df = pd.DataFrame(columns=['ObjectURL', 'Size'])
+        for object in bucket_objects:
+            result_objects_df = result_objects_df.append([{'ObjectURL': object['Key'], 'Size': object['Size']}], ignore_index=True)
+
+        return result_objects_df
+
     def delete_result(self, jobId):
         if not self.logged_on:
             print("You are not logged on to IBM Cloud")
@@ -205,6 +305,8 @@ class SQLQuery():
             raise ValueError('SQL job with jobId {} did not finish successfully. No result available.')
 
         result_location = job_details['resultset_location'].replace("cos", "https", 1)
+        provided_cos_endpoint = job_details['resultset_location'].split("/")[2]
+        result_cos_endpoint = self.endpoint_alias_mapping.get(provided_cos_endpoint, provided_cos_endpoint)
 
         fourth_slash = result_location.replace('/', 'X', 3).find('/')
 
@@ -235,7 +337,7 @@ class SQLQuery():
                                       ibm_api_key_id=self.api_key,
                                       ibm_auth_endpoint="https://iam.ng.bluemix.net/oidc/token",
                                       config=Config(signature_version='oauth'),
-                                      endpoint_url='https://' + self.target_cos_endpoint)
+                                      endpoint_url='https://' + result_cos_endpoint)
 
         response = cos_client.delete_objects(Bucket=bucket_name, Delete={'Objects': bucket_objects})
 
@@ -252,7 +354,7 @@ class SQLQuery():
 
         try:
             response = self.client.fetch(
-                "https://sql-api.ng.bluemix.net/v2-beta/sql_jobs/{}?instance_crn={}".format(jobId, self.instance_crn),
+                "https://sql-api.ng.bluemix.net/v2/sql_jobs/{}?instance_crn={}".format(jobId, self.instance_crn),
                 method='GET',
                 headers=self.request_headers,
                 validate_cert=False)
@@ -270,17 +372,18 @@ class SQLQuery():
             return
 
         response = self.client.fetch(
-            "https://sql-api.ng.bluemix.net/v2-beta/sql_jobs?instance_crn={}".format(self.instance_crn),
+            "https://sql-api.ng.bluemix.net/v2/sql_jobs?instance_crn={}".format(self.instance_crn),
             method='GET',
             headers=self.request_headers,
             validate_cert=False)
         if response.code == 200 or response.code == 201:
             job_list = json_decode(response.body)
             job_list_df = pd.DataFrame(columns=['job_id', 'status', 'user_id', 'statement', 'resultset_location',
-                                                'submit_time', 'end_time', 'error', 'error_message'])
+                                                'submit_time', 'end_time', 'rows_read', 'rows_returned', 'bytes_read',
+                                                'error', 'error_message'])
             for job in job_list['jobs']:
                 response = self.client.fetch(
-                    "https://sql-api.ng.bluemix.net/v2-beta/sql_jobs/{}?instance_crn={}".format(job['job_id'],
+                    "https://sql-api.ng.bluemix.net/v2/sql_jobs/{}?instance_crn={}".format(job['job_id'],
                                                                                                 self.instance_crn),
                     method='GET',
                     headers=self.request_headers,
@@ -288,14 +391,23 @@ class SQLQuery():
                 if response.code == 200 or response.code == 201:
                     job_details = json_decode(response.body)
                     error = None
+                    error_message = None
+                    rows_read = None
+                    rows_returned = None
+                    bytes_read = None
+                    end_time = None
                     if 'error' in job_details:
                         error = job_details['error']
-                        end_time = None
-                    else:
+                    if 'end_time' in job_details:
                         end_time = job_details['end_time']
-                    error_message = None
                     if 'error_message' in job_details:
                         error_message = job_details['error_message']
+                    if 'rows_read' in job_details:
+                        rows_read = job_details['rows_read']
+                    if 'rows_returned' in job_details:
+                        rows_returned = job_details['rows_returned']
+                    if 'bytes_read' in job_details:
+                        bytes_read = job_details['bytes_read']
                     job_list_df = job_list_df.append([{'job_id': job['job_id'],
                                                        'status': job_details['status'],
                                                        'user_id': job_details['user_id'],
@@ -303,6 +415,9 @@ class SQLQuery():
                                                        'resultset_location': job_details['resultset_location'],
                                                        'submit_time': job_details['submit_time'],
                                                        'end_time': end_time,
+                                                       'rows_read': rows_read,
+                                                       'rows_returned': rows_returned,
+                                                       'bytes_read': bytes_read,
                                                        'error': error,
                                                        'error_message': error_message,
                                                        }], ignore_index=True)
@@ -405,3 +520,54 @@ class SQLQuery():
                 'newest_object_timestamp': newest_modification,
                 'smallest_object_size': sizeof_fmt(smallest_size), 'smallest_object': smallest_object,
                 'largest_object_size': sizeof_fmt(largest_size), 'largest_object': largest_object}
+
+    def export_job_history(self, cos_url=None):
+        if cos_url:
+            # Default export location is target COS URL set at __init__
+            # But we'll overwrite that with the provided export URL
+            self.export_cos_url = cos_url
+        elif not self.export_cos_url:
+            raise ValueError('No configured export COS URL.')
+        if not self.export_cos_url.endswith('/'):
+            self.export_cos_url += "/"
+        provided_export_cos_endpoint = self.export_cos_url.split("/")[2]
+        export_cos_endpoint = self.endpoint_alias_mapping.get(provided_export_cos_endpoint, provided_export_cos_endpoint)
+        export_cos_bucket = self.export_cos_url.split("/")[3]
+        export_cos_prefix = self.export_cos_url[self.export_cos_url.replace('/', 'X', 3).find('/')+1:]
+        export_file_prefix = "job_export_"
+
+        job_history_df = self.get_jobs() # Retrieve current job history (most recent 30 jobs)
+        terminated_job_history_df = job_history_df[job_history_df['status'].isin(['completed', 'failed'])] # Only export terminated jobs
+        newest_job_end_time = terminated_job_history_df.loc[pd.to_datetime(terminated_job_history_df['end_time']).idxmax()].end_time
+
+        # List all existing objects in export location and identify latest exported job timestamp:
+        cos_client = ibm_boto3.client(service_name='s3',
+                                      ibm_api_key_id=self.api_key,
+                                      ibm_auth_endpoint="https://iam.ng.bluemix.net/oidc/token",
+                                      config=Config(signature_version='oauth'),
+                                      endpoint_url='https://' + export_cos_endpoint)
+        paginator = cos_client.get_paginator("list_objects")
+        page_iterator = paginator.paginate(Bucket=export_cos_bucket, Prefix=export_cos_prefix)
+        newest_exported_job_end_time = ""
+        for page in page_iterator:
+            if "Contents" in page:
+                for key in page['Contents']:
+                    object_name = key["Key"]
+                    suffix_index = object_name.find(".parquet")
+                    prefix_end_index = len(export_cos_prefix + export_file_prefix)
+                    if prefix_end_index < suffix_index:
+                        job_end_time = object_name[prefix_end_index:suffix_index]
+                        if job_end_time > newest_exported_job_end_time:
+                            newest_exported_job_end_time = job_end_time
+
+        # Export all new jobs if there are some:
+        if newest_exported_job_end_time < newest_job_end_time:
+            tmpfile = tempfile.NamedTemporaryFile()
+            tempfilename = tmpfile.name
+            new_jobs_df = terminated_job_history_df[terminated_job_history_df['end_time'] > newest_exported_job_end_time]
+            new_jobs_df.to_parquet(engine="pyarrow", fname=tempfilename, compression="snappy")
+            cos_client.upload_file(Bucket=export_cos_bucket, Filename=tempfilename, Key=export_cos_prefix + export_file_prefix + newest_job_end_time + ".parquet")
+            print("Exported {} new jobs".format(new_jobs_df['job_id'].count()))
+            tmpfile.close()
+        else:
+            print("No new jobs to export")
\ No newline at end of file
